
/******************************************************************************
* MODULE     : simplify.cc
* DESCRIPTION: Rudimentary simplification of expression used before printing
*              Not much should be done so that the original intention of
*              the way things should be output is kept
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <simplify.hh>
#include <list.hh>
#include <symbols.hh>
#include <literal.hh>
namespace mmx {

static symbol Zero ("0");
static symbol One  ("1");

inline bool
is_neg (generic g) {
  return is_func (g, MMX_MINUS, 2) && (g[1] == Zero);
}

inline bool
is_inv (generic g) {
  return is_func (g, MMX_OVER, 2) && (g[1] == One);
}

static generic simp (generic g);
static generic pre_simp (generic g);
static generic post_simp (generic g);

/******************************************************************************
* Pretreatment and posttreatment
******************************************************************************/

static list
pre_simp (list l) {
  if (nil (l)) return l;
  return cons (pre_simp (car (l)), pre_simp (cdr (l)));
}

static generic
pre_simp (generic g) {
  if (is_literal (g)) return pre_simp (S(g));
  if (is_symbol (g) && is_string (S (as_symbol (g)))) {
    string s= as_string (S (as_symbol (g)));
    if ((N(s) >= 2) && (s[0]=='-'))
      if ((s[1] == '.') || ((s[1] >= '0') && (s[1] <= '9')))
	return Minus (Zero, symbol (s (1, N(s))));
    return g;
  }
  if ((!is_list (g)) || nil (g)) return g;
  list ops= cdr (as_list (g));
  if (is_func (g, MMX_PLUS, 0)) return Zero;
  if (is_func (g, MMX_TIMES, 0)) return One;
  if (is_func (g, MMX_PLUS) && (N(g) > 3))
    return Plus (g[1], pre_simp (cons (MMX_PLUS, cdr (ops))));
  if (is_func (g, MMX_MINUS) && (N(g) > 3))
    return Minus (g[1], pre_simp (cons (MMX_PLUS, cdr (ops))));
  if (is_func (g, MMX_TIMES) && (N(g) > 3))
    return Times (g[1], pre_simp (cons (MMX_TIMES, cdr (ops))));
  if (is_func (g, MMX_OVER) && (N(g) > 3))
    return Over (g[1], pre_simp (cons (MMX_TIMES, cdr (ops))));
  g= cons (car (g), pre_simp (ops));
  if (is_func (g, MMX_PLUS, 1)) return g[1];
  if (is_func (g, MMX_MINUS, 1)) return Minus (Zero, g[1]);
  if (is_func (g, MMX_TIMES, 1)) return g[1];
  if (is_func (g, MMX_OVER, 1)) return Over (One, g[1]);
  return g;
}

static list
post_simp (list l) {
  if (nil (l)) return l;
  return cons (post_simp (car (l)), post_simp (cdr (l)));
}

static generic
post_simp (generic g) {
  if ((!is_list (g)) || nil (g)) return g;
  // These rules are superfluous if * and / take precendence over unary minary
  if (is_neg (g) && is_func (g[2], MMX_TIMES, 2))
    return post_simp (Times (Minus (Zero, g[2][1]), g[2][2]));
  if (is_neg (g) && is_func (g[2], MMX_OVER, 2))
    return post_simp (Over (Minus (Zero, g[2][1]), g[2][2]));
  // End of rules -(ab) -> (-a)b and -(a/b) -> (-a)/b
  g= cons (car (g), post_simp (cdr (as_list (g))));
  if (is_neg (g)) return Minus (g[2]);
  return g;
}

/******************************************************************************
* Main simplification routines
******************************************************************************/

static list
simp (list l) {
  if (nil (l)) return l;
  return cons (simp (car (l)), simp (cdr (l)));
}

static generic
simp (generic g) {
  if (!is_list (g)) return g;
  if (N(g) <= 1) return g;

  if (is_func (g, MMX_PLUS, 2) && is_func (g[2], MMX_PLUS, 2))
    return simp (Plus (Plus (g[1], g[2][1]), g[2][2]));
  if (is_func (g, MMX_PLUS, 2) && is_func (g[2], MMX_MINUS, 2))
    return simp (Minus (Plus (g[1], g[2][1]), g[2][2]));
  if (is_func (g, MMX_MINUS, 2) && is_func (g[2], MMX_PLUS, 2))
    return simp (Minus (Minus (g[1], g[2][1]), g[2][2]));
  if (is_func (g, MMX_MINUS, 2) && is_func (g[2], MMX_MINUS, 2))
    return simp (Plus (Minus (g[1], g[2][1]), g[2][2]));

  if (is_func (g, MMX_TIMES, 2) && is_func (g[2], MMX_TIMES, 2))
    return simp (Times (Times (g[1], g[2][1]), g[2][2]));
  if (is_func (g, MMX_TIMES, 2) && is_func (g[2], MMX_OVER, 2))
    return simp (Over (Times (g[1], g[2][1]), g[2][2]));
  if (is_func (g, MMX_OVER, 2) && is_func (g[2], MMX_TIMES, 2))
    return simp (Over (Over (g[1], g[2][1]), g[2][2]));
  if (is_func (g, MMX_OVER, 2) && is_func (g[2], MMX_OVER, 2))
    return simp (Times (Over (g[1], g[2][1]), g[2][2]));

  g= cons (car (g), simp (cdr (as_list (g))));

  if (is_func (g, MMX_POWER, 2) && (g[1] == Zero)) return One;
  if (is_func (g, MMX_POWER, 2) && (g[1] == One)) return g[1];

  if (is_func (g, MMX_TIMES, 2) && (g[1] == One)) return g[2];
  if (is_func (g, MMX_TIMES, 2) && (g[2] == One)) return g[1];
  if (is_func (g, MMX_OVER, 2) && (g[2] == One)) return g[1];
  if (is_func (g, MMX_OVER, 2) && (g[1] == g[2])) return One;
  if (is_inv (g) && is_inv (g[2])) return g[2][2];

  if (is_func (g, MMX_TIMES, 2) && (g[1] == Zero)) return Zero;
  if (is_func (g, MMX_TIMES, 2) && (g[2] == Zero)) return Zero;
  if (is_func (g, MMX_TIMES, 2) && is_neg (g[1]) && is_neg (g[2]))
    return Times (g[1][2], g[2][2]);
  if (is_func (g, MMX_TIMES, 2) && is_neg (g[1]))
    return Minus (Zero, Times (g[1][2], g[2]));
  if (is_func (g, MMX_TIMES, 2) && is_neg (g[2]))
    return Minus (Zero, Times (g[1], g[2][2]));
  if (is_func (g, MMX_OVER, 2) && (g[1] == Zero)) return Zero;
  if (is_func (g, MMX_OVER, 2) && is_neg (g[1]) && is_neg (g[2]))
    return Over (g[1][2], g[2][2]);
  if (is_func (g, MMX_OVER, 2) && is_neg (g[1]))
    return Minus (Zero, Over (g[1][2], g[2]));
  if (is_func (g, MMX_OVER, 2) && is_neg (g[2]))
    return Minus (Zero, Over (g[1], g[2][2]));

  if (is_func (g, MMX_PLUS, 2) && (g[1] == Zero)) return g[2];
  if (is_func (g, MMX_PLUS, 2) && (g[2] == Zero)) return g[1];
  if (is_func (g, MMX_MINUS, 2) && (g[2] == Zero)) return g[1];
  if (is_func (g, MMX_MINUS, 2) && (g[1] == g[2])) return Zero;
  if (is_neg (g) && is_neg (g[2])) return g[2][2];

  return g;
}

/******************************************************************************
* User interface
******************************************************************************/

generic
simplify (generic g1) {
  //cout << "g1= " << g1 << lf;
  generic g2= pre_simp (g1);
  //cout << "g2= " << g2 << lf;
  generic g3;
  while (g3 != g2) {
    g3= g2;
    g2= simp (g2);
  }
  //cout << "g3= " << g3 << lf;
  generic g4= post_simp (g3);
  //cout << "g4= " << g4 << lf;
  return g4;
}

} // namespace mmx
